import unittest
from datetime import timedelta

from biogeme_optimization.pareto import Pareto

from biogeme.results_processing import (
    EstimationResults,
    RawEstimationResults,
    pareto_optimal,
)


class TestParetoOptimal(unittest.TestCase):

    def setUp(self):
        # Set up multiple real EstimationResults for the Pareto optimal test
        raw_estimation_1 = RawEstimationResults(
            model_name='model_1',
            user_notes='user notes 1',
            beta_names=['beta1', 'beta2'],
            beta_values=[1.2, 1.3],
            lower_bounds=[-10, -11],
            upper_bounds=[None, 23],
            gradient=[0.001, -0.002],
            hessian=[[1, 2], [3, 4]],
            bhhh=[[10, 20], [30, 40]],
            null_log_likelihood=-1000,
            initial_log_likelihood=-2000,
            final_log_likelihood=-500,
            data_name='data_1',
            sample_size=1000,
            number_of_observations=900,
            monte_carlo=True,
            number_of_draws=5000,
            types_of_draws={'var1': 'UNIFORM'},
            number_of_excluded_data=10,
            draws_processing_time=timedelta(seconds=15),
            optimization_messages={'Diagnostic': 'optimal'},
            convergence=True,
            bootstrap=[[1.0, 1.0], [2.0, 2.0]],
            bootstrap_time=timedelta(seconds=10),
        )

        raw_estimation_2 = RawEstimationResults(
            model_name='model_2',
            user_notes='user notes 2',
            beta_names=['beta3', 'beta4'],
            beta_values=[2.1, 2.4],
            lower_bounds=[-5, -6],
            upper_bounds=[None, 20],
            gradient=[0.002, -0.003],
            hessian=[[5, 6], [7, 8]],
            bhhh=[[15, 25], [35, 45]],
            null_log_likelihood=-900,
            initial_log_likelihood=-1800,
            final_log_likelihood=-700,
            data_name='data_2',
            sample_size=1100,
            number_of_observations=950,
            monte_carlo=True,
            number_of_draws=6000,
            types_of_draws={'var2': 'NORMAL'},
            number_of_excluded_data=15,
            draws_processing_time=timedelta(seconds=20),
            optimization_messages={'Diagnostic': 'converged'},
            convergence=True,
            bootstrap=[[2.0, 2.0], [3.0, 3.0]],
            bootstrap_time=timedelta(seconds=12),
        )

        # Create estimation results objects
        self.estimation_results_1 = EstimationResults(
            raw_estimation_results=raw_estimation_1
        )
        self.estimation_results_2 = EstimationResults(
            raw_estimation_results=raw_estimation_2
        )

        # Dictionary of results to test Pareto optimality
        self.dict_of_results = {
            'config_1': self.estimation_results_1,
            'config_2': self.estimation_results_2,
        }

    def test_pareto_optimal(self):
        # Call the function to identify Pareto optimal models
        optimal_results = pareto_optimal(self.dict_of_results)

        # Check if the result is a dictionary
        self.assertIsInstance(optimal_results, dict)

        # Ensure the results contain only Pareto optimal configurations
        # Based on the final log-likelihood and number of parameters, only one should be Pareto optimal
        # Model 1 has a better log-likelihood (-500), while Model 2 has more parameters
        self.assertIn('config_1', optimal_results)  # config_1 should be included
        self.assertNotIn('config_2', optimal_results)  # config_2 is not Pareto optimal

        # Validate the content of the Pareto optimal result
        self.assertEqual(optimal_results['config_1'], self.estimation_results_1)

    def test_pareto_with_existing_pareto(self):
        # Test with an existing Pareto set passed in
        existing_pareto = Pareto()
        optimal_results = pareto_optimal(self.dict_of_results, a_pareto=existing_pareto)

        # Ensure the results still reflect the correct Pareto optimal configuration
        self.assertIn('config_1', optimal_results)
        self.assertNotIn('config_2', optimal_results)

        # Ensure that the Pareto set was modified
        self.assertTrue(len(existing_pareto.pareto) > 0)

    def test_pareto_empty(self):
        # Test with an empty dictionary
        empty_dict = {}
        optimal_results = pareto_optimal(empty_dict)

        # Ensure that no results are returned for an empty input
        self.assertEqual(len(optimal_results), 0)


if __name__ == '__main__':
    unittest.main()



import unittest

import numpy as np
from biogeme.calculator import create_function_simple_expression
from biogeme.expressions import Beta, log, logzero
from biogeme.function_output import FunctionOutput
from biogeme.likelihood.negative_likelihood import NegativeLikelihood


class test_expressions(unittest.TestCase):
    def test_log(self):
        argument = Beta('argument', 2, None, None, 0)
        the_log = log(argument)
        expression_function = create_function_simple_expression(
            expression=the_log, numerically_safe=True
        )
        the_function_output: FunctionOutput = expression_function(
            [10.0], gradient=False, hessian=False, bhhh=False
        )
        self.assertAlmostEqual(the_function_output.function, np.log(10), 3)
        optimization_function = NegativeLikelihood(
            dimension=1,
            loglikelihood=create_function_simple_expression(
                expression=the_log, numerically_safe=True
            ),
        )
        check_results = optimization_function.check_derivatives(x=[1.0])
        for a_grad, fd_grad in zip(
            check_results.analytical.gradient, check_results.finite_differences.gradient
        ):
            self.assertAlmostEqual(a_grad, fd_grad, places=2)
        for a_hess, fd_hess in zip(
            np.nditer(check_results.analytical.hessian),
            np.nditer(check_results.finite_differences.hessian),
        ):
            self.assertAlmostEqual(a_hess, fd_hess, delta=1.0e-4)

        # Log of a negative number
        the_function_output: FunctionOutput = expression_function(
            [-1.0], gradient=True, hessian=False, bhhh=False
        )
        self.assertTrue(the_function_output.function < -1e10)
        self.assertGreater(the_function_output.gradient[0], 0.0)
        optimization_function = NegativeLikelihood(
            dimension=1,
            loglikelihood=create_function_simple_expression(
                expression=the_log, numerically_safe=True
            ),
        )
        check_results = optimization_function.check_derivatives(x=[-1.0])
        for a_grad, fd_grad in zip(
            check_results.analytical.gradient, check_results.finite_differences.gradient
        ):
            self.assertAlmostEqual(a_grad, fd_grad, places=2)
        for a_hess, fd_hess in zip(
            np.nditer(check_results.analytical.hessian),
            np.nditer(check_results.finite_differences.hessian),
        ):
            self.assertAlmostEqual(a_hess, fd_hess, delta=1.0e-4)

        # Log of zero
        the_function_output: FunctionOutput = expression_function(
            [0.0], gradient=True, hessian=False, bhhh=False
        )
        self.assertTrue(the_function_output.function < -10)
        self.assertGreater(the_function_output.gradient[0], 0.0)
        optimization_function = NegativeLikelihood(
            dimension=1,
            loglikelihood=create_function_simple_expression(
                expression=the_log, numerically_safe=True
            ),
        )
        check_results = optimization_function.check_derivatives(x=[-1.0])
        for a_grad, fd_grad in zip(
            check_results.analytical.gradient,
            check_results.finite_differences.gradient,
        ):
            self.assertAlmostEqual(a_grad, fd_grad, places=2)
        for a_hess, fd_hess in zip(
            np.nditer(check_results.analytical.hessian),
            np.nditer(check_results.finite_differences.hessian),
        ):
            self.assertAlmostEqual(a_hess, fd_hess, delta=1.0e-4)

    def test_logzero(self):
        argument = Beta('argument', 2, None, None, 0)
        the_log = logzero(argument)
        expression_function = create_function_simple_expression(
            expression=the_log, numerically_safe=True
        )
        the_function_output: FunctionOutput = expression_function(
            [10.0], gradient=False, hessian=False, bhhh=False
        )
        self.assertAlmostEqual(the_function_output.function, np.log(10), 3)
        optimization_function = NegativeLikelihood(
            dimension=1,
            loglikelihood=create_function_simple_expression(
                expression=the_log, numerically_safe=True
            ),
        )
        check_results = optimization_function.check_derivatives(x=[1.0])
        for a_grad, fd_grad in zip(
            check_results.analytical.gradient, check_results.finite_differences.gradient
        ):
            self.assertAlmostEqual(a_grad, fd_grad, places=2)
        for a_hess, fd_hess in zip(
            np.nditer(check_results.analytical.hessian),
            np.nditer(check_results.finite_differences.hessian),
        ):
            self.assertAlmostEqual(a_hess, fd_hess, delta=1.0e-4)

        # Log of a negative number
        the_function_output: FunctionOutput = expression_function(
            [-1.0], gradient=True, hessian=False, bhhh=False
        )
        self.assertTrue(the_function_output.function < -1e10)
        self.assertGreater(the_function_output.gradient[0], 0.0)
        optimization_function = NegativeLikelihood(
            dimension=1,
            loglikelihood=create_function_simple_expression(
                expression=the_log, numerically_safe=True
            ),
        )
        check_results = optimization_function.check_derivatives(x=[-1.0])
        for a_grad, fd_grad in zip(
            check_results.analytical.gradient, check_results.finite_differences.gradient
        ):
            self.assertAlmostEqual(a_grad, fd_grad, places=2)
        for a_hess, fd_hess in zip(
            np.nditer(check_results.analytical.hessian),
            np.nditer(check_results.finite_differences.hessian),
        ):
            self.assertAlmostEqual(a_hess, fd_hess, delta=1.0e-4)

        # Log of zero
        the_function_output: FunctionOutput = expression_function(
            [0.0], gradient=True, hessian=False, bhhh=False
        )
        self.assertEqual(the_function_output.function, 0)
        self.assertEqual(the_function_output.gradient[0], 0.0)
        optimization_function = NegativeLikelihood(
            dimension=1,
            loglikelihood=create_function_simple_expression(
                expression=the_log, numerically_safe=True
            ),
        )
        check_results = optimization_function.check_derivatives(x=[-1.0])
        for a_grad, fd_grad in zip(
            check_results.analytical.gradient,
            check_results.finite_differences.gradient,
        ):
            self.assertAlmostEqual(a_grad, fd_grad, places=2)
        for a_hess, fd_hess in zip(
            np.nditer(check_results.analytical.hessian),
            np.nditer(check_results.finite_differences.hessian),
        ):
            self.assertAlmostEqual(a_hess, fd_hess, delta=1.0e-4)

    def test_divide(self):
        numerator = Beta('numerator', 1, None, None, 0)
        denominator = Beta('denominator', 1, None, None, 0)
        ratio = numerator / denominator
        expression_function = create_function_simple_expression(
            expression=ratio, numerically_safe=True
        )
        the_function_output: FunctionOutput = expression_function(
            [1, 1], gradient=True, hessian=False, bhhh=False
        )
        expected_f = 1
        expected_g = [1, -1]
        self.assertAlmostEqual(the_function_output.function, expected_f, 3)
        self.assertAlmostEqual(the_function_output.gradient[0], expected_g[0], 3)
        self.assertAlmostEqual(the_function_output.gradient[1], expected_g[1], 3)

        # Division by zero
        the_function_output: FunctionOutput = expression_function(
            [1, 0], gradient=True, hessian=False, bhhh=False
        )
        expected_f = 0
        expected_g = [0, 4503599627370496.0]
        self.assertAlmostEqual(the_function_output.function, expected_f, 3)
        self.assertAlmostEqual(the_function_output.gradient[0], expected_g[0], 3)
        self.assertAlmostEqual(
            the_function_output.gradient[1], expected_g[1], delta=1e2
        )

    def test_power(self):
        base = Beta('base', 1, None, None, 0)
        exponent = Beta('exponent', 1, None, None, 0)
        power = base**exponent
        expression_function = create_function_simple_expression(
            expression=power, numerically_safe=True
        )
        the_function_output: FunctionOutput = expression_function(
            [1, 1], gradient=True, hessian=False, bhhh=False
        )
        expected_f = 1
        expected_g = [1, 0]
        self.assertAlmostEqual(the_function_output.function, expected_f, 3)
        self.assertAlmostEqual(the_function_output.gradient[0], expected_g[0], 3)
        self.assertAlmostEqual(the_function_output.gradient[1], expected_g[1], 3)

        optimization_function = NegativeLikelihood(
            dimension=1,
            loglikelihood=create_function_simple_expression(
                expression=power, numerically_safe=True
            ),
        )
        check_results = optimization_function.check_derivatives(x=[1.0, 1.0])
        for a_grad, fd_grad in zip(
            check_results.analytical.gradient,
            check_results.finite_differences.gradient,
        ):
            self.assertAlmostEqual(a_grad, fd_grad, places=2)
        for a_hess, fd_hess in zip(
            np.nditer(check_results.analytical.hessian),
            np.nditer(check_results.finite_differences.hessian),
        ):
            self.assertAlmostEqual(a_hess, fd_hess, delta=1.0e-4)

        base_value = 2.0
        exponent_value = 3.0
        base = Beta('base', base_value, None, None, 0)
        exponent = Beta('exponent', exponent_value, None, None, 0)
        power = base**exponent
        expression_function = create_function_simple_expression(
            expression=power, numerically_safe=True
        )
        the_function_output: FunctionOutput = expression_function(
            [base_value, exponent_value], gradient=True, hessian=False, bhhh=False
        )
        expected_f = base_value**exponent_value
        expected_g = [12, 5.545]
        self.assertAlmostEqual(the_function_output.function, expected_f, 3)
        self.assertAlmostEqual(the_function_output.gradient[0], expected_g[0], 3)
        self.assertAlmostEqual(the_function_output.gradient[1], expected_g[1], 3)

        base_value = 0.0
        exponent_value = 0.0
        base = Beta('base', base_value, None, None, 0)
        exponent = Beta('exponent', exponent_value, None, None, 0)
        power = base**exponent
        expression_function = create_function_simple_expression(
            expression=power, numerically_safe=True
        )
        the_function_output: FunctionOutput = expression_function(
            [base_value, exponent_value], gradient=True, hessian=False, bhhh=False
        )
        expected_f = 0
        expected_g = [0, 0]
        self.assertAlmostEqual(the_function_output.function, expected_f, 3)
        self.assertAlmostEqual(the_function_output.gradient[0], expected_g[0], 3)
        self.assertAlmostEqual(the_function_output.gradient[1], expected_g[1], 3)

        base_value = 1.0e-6
        exponent_value = -2
        base = Beta('base', base_value, None, None, 0)
        exponent = Beta('exponent', exponent_value, None, None, 0)
        power = base**exponent
        expression_function = create_function_simple_expression(
            expression=power, numerically_safe=True
        )
        the_function_output: FunctionOutput = expression_function(
            [base_value, exponent_value], gradient=True, hessian=False, bhhh=False
        )
        expected_f = 1.0e12
        expected_g = [0, 0]
        self.assertAlmostEqual(the_function_output.function, expected_f, 2)
        self.assertTrue(the_function_output.gradient[0] < 0)
        self.assertTrue(the_function_output.gradient[1] < 0)



"""Define various items for Jax

Michel Bierlaire
Tue Mar 18 18:28:07 2025
"""

from collections.abc import Callable

import jax.numpy as jnp
from jax import jit, vmap

JaxFunctionType = Callable[
    [jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray], jnp.array
]


def build_vectorized_function(the_function):
    """Build the function that is applied to each row of the databaser"""

    def vectorized_function(parameters, data, draws, random_variables):
        return vmap(
            lambda row, draw: the_function(parameters, row, draw, random_variables),
            in_axes=(0, 0),
        )(data, draws)

    return jit(vectorized_function)



"""

biogeme.distributions
=====================

Example of usage of the distributions module.  This is for programmers who need
examples of use of the functions of the class. The examples are
designed to illustrate the syntax.

:author: Michel Bierlaire
:date: Fri Nov 17 08:27:24 2023
"""

import biogeme.version as ver
import biogeme.distributions as dist
from biogeme.expressions import Beta

print(ver.getText())

# %%
# pdf of the normal distributio: returns the biogeme expression of the
# probability density function of the normal distribution:
#
# .. math:: f(x;\mu, \sigma) = \frac{1}{\sigma \sqrt{2\pi}}
#           \exp{-\frac{(x-\mu)^2}{2\sigma^2}}.

# %%
# Calculated for a numeric value.
resulting_expression = dist.normalpdf(0)
resulting_expression

# %%
resulting_expression.getValue()

# %%
# Calculated for an expression.
a_parameter = Beta('a_parameter', 0, None, None, 1)
mu = Beta('mu', 0, None, None, 1)
sigma = Beta('sigma', 1, None, None, 1)

# %%
resulting_expression = dist.normalpdf(a_parameter, mu=mu, s=sigma)
resulting_expression

# %%
resulting_expression.getValue()

# %%
# pdf of the lognormal distribution: returns the biogeme expression of
# the probability density function of the lognormal distribution
#
# .. math:: f(x;\mu, \sigma) = \frac{1}{x\sigma \sqrt{2\pi}}
#               \exp{-\frac{(\ln x-\mu)^2}{2\sigma^2}}

# %%
# Calculated for a numeric value.
resulting_expression = dist.lognormalpdf(1)
resulting_expression

# %%
resulting_expression.getValue()

# %%
# Calculated for an expression.
a_parameter = Beta('a_parameter', 1, None, None, 1)
mu = Beta('mu', 0, None, None, 1)
sigma = Beta('sigma', 1, None, None, 1)

# %%
resulting_expression = dist.lognormalpdf(a_parameter, mu=mu, s=sigma)
resulting_expression

# %%
resulting_expression.getValue()

# %%
# pdf of the uniform distribution: returns the biogeme expression of
# the probability density function of the uniform distribution
#
# .. math:: f(x; a, b) = \left\{ \begin{array}{ll}
#               \frac{1}{b-a} & \mbox{for } x \in [a, b] \\
#               0 & \mbox{otherwise}\end{array} \right.

# %%
# Calculated for a numeric value
resulting_expression = dist.uniformpdf(0)
resulting_expression

# %%
resulting_expression.getValue()

# %%
# Calculated for an expression
a_parameter = Beta('a_parameter', 0, None, None, 1)
a = Beta('a', -1, None, None, 1)
b = Beta('b', 1, None, None, 1)

# %%
resulting_expression = dist.uniformpdf(a_parameter, a=a, b=b)
resulting_expression

# %%
resulting_expression.getValue()

# %%
# pdf of the triangular distribution: returns the biogeme expression
# of the probability density function of the triangular distribution
#
#
# .. math:: f(x;a, b, c) = \left\{ \begin{array}{ll} 0 &
#              \text{if } x < a \\\frac{2(x-a)}{(b-a)(c-a)} &
#              \text{if } a \leq x < c \\\frac{2(b-x)}{(b-a)(b-c)} &
#              \text{if } c \leq x < b \\0 & \text{if } x \geq b.
#              \end{array} \right.
#
# It is assumed that :math:`a < c < b`. It is not verified.

# %%
# Calculated for a numeric value.
resulting_expression = dist.triangularpdf(0)
resulting_expression

# %%
resulting_expression.getValue()

# %%
# Calculated for an expression.
a_parameter = Beta('a_parameter', 0, None, None, 1)
a = Beta('a', -1, None, None, 1)
b = Beta('b', 1, None, None, 1)
c = Beta('c', 0, None, None, 1)

# %%
resulting_expression = dist.triangularpdf(a_parameter, a=a, b=b, c=c)
resulting_expression

# %%
resulting_expression.getValue()

# %%
# CDF of the logistic distribution: returns the biogeme expression of
# the cumulative distribution function of the logistic distribution
#
#
# .. math:: f(x;\mu, \sigma) = \frac{1}{1+\exp\left(-\frac{x-\mu}{\sigma} \right)}
#
#

# %%
# Calculated for a numeric value
resulting_expression = dist.logisticcdf(0)
resulting_expression

# %%
resulting_expression.getValue()

# %%
# Calculated for an expression
a_parameter = Beta('a_parameter', 0, None, None, 1)
mu = Beta('mu', 0, None, None, 1)
sigma = Beta('sigma', 1, None, None, 1)

# %%
resulting_expression = dist.logisticcdf(a_parameter, mu=mu, s=sigma)
resulting_expression

# %%
resulting_expression.getValue()



""".. _plot_b21multiple_models_spec:

Specification of a catalog of models
====================================

Specification of the catalogs used by the assisted specification
algorithm. Note that this script does not perform any estimation. It
is imported by other scripts: :ref:`plot_b21multiple_models`, :ref:`plot_b21process_pareto`.

:author: Michel Bierlaire, EPFL
:date: Fri Jul 21 17:46:09 2023

"""
import biogeme.biogeme as bio
from biogeme import models
from biogeme.expressions import Beta, logzero
from biogeme.catalog import Catalog, segmentation_catalogs

# %%
# See the data processing script: :ref:`swissmetro_data`.
from swissmetro_data import (
    database,
    CHOICE,
    SM_AV,
    CAR_AV_SP,
    TRAIN_AV_SP,
    TRAIN_TT_SCALED,
    TRAIN_COST_SCALED,
    SM_TT_SCALED,
    SM_COST_SCALED,
    CAR_TT_SCALED,
    CAR_CO_SCALED,
    MALE,
    INCOME,
    GA,
)

# %%
# Parameters to be estimated.
ASC_CAR = Beta('ASC_CAR', 0, None, None, 0)
ASC_TRAIN = Beta('ASC_TRAIN', 0, None, None, 0)
B_TIME = Beta('B_TIME', 0, None, None, 0)
B_COST = Beta('B_COST', 0, None, None, 0)

# %%
# Segmentations.
gender_segmentation = database.generate_segmentation(
    variable=MALE,
    mapping={
        0: 'female',
        1: 'male',
    },
)

# %%
income_segmentation = database.generate_segmentation(
    variable=INCOME,
    mapping={
        0: 'inc-zero',
        1: 'inc-under50',
        2: 'inc-50-100',
        3: 'inc-100+',
        4: 'inc-unknown',
    },
)

print(f'{income_segmentation=}')

# %%
ga_segmentation = database.generate_segmentation(
    variable=GA, mapping={1: 'GA', 0: 'noGA'}
)

# %%
asc_segmentations = (
    gender_segmentation,
    ga_segmentation,
)

# %%
ASC_CAR_catalog, ASC_TRAIN_catalog = segmentation_catalogs(
    generic_name='ASC',
    beta_parameters=[ASC_CAR, ASC_TRAIN],
    potential_segmentations=asc_segmentations,
    maximum_number=2,
)

# %%
cost_segmentations = (
    ga_segmentation,
    income_segmentation,
)

# %%
# Note that the function returns a list. In this case, it contains
# only one element. This is the reason of the presence of the comma
# after B_COST_catalog
(B_COST_catalog,) = segmentation_catalogs(
    generic_name='B_COST',
    beta_parameters=[B_COST],
    potential_segmentations=cost_segmentations,
    maximum_number=1,
)

# %%
# Parameter for Box-Cox transforms
ell_time = Beta('lambda_time', 1, None, 10, 0)

# %%
# Potential non linear specifications of travel time.
TRAIN_TT_catalog = Catalog.from_dict(
    catalog_name='TRAIN_TT',
    dict_of_expressions={
        'linear': TRAIN_TT_SCALED,
        'log': logzero(TRAIN_TT_SCALED),
        'boxcox': models.boxcox(TRAIN_TT_SCALED, ell_time),
    },
)

SM_TT_catalog = Catalog.from_dict(
    catalog_name='SM_TT',
    dict_of_expressions={
        'linear': SM_TT_SCALED,
        'log': logzero(SM_TT_SCALED),
        'boxcox': models.boxcox(SM_TT_SCALED, ell_time),
    },
    controlled_by=TRAIN_TT_catalog.controlled_by,
)

CAR_TT_catalog = Catalog.from_dict(
    catalog_name='CAR_TT',
    dict_of_expressions={
        'linear': CAR_TT_SCALED,
        'log': logzero(CAR_TT_SCALED),
        'boxcox': models.boxcox(CAR_TT_SCALED, ell_time),
    },
    controlled_by=TRAIN_TT_catalog.controlled_by,
)

# %%
# Definition of the utility functions with linear cost.
V1 = ASC_TRAIN_catalog + B_TIME * TRAIN_TT_catalog + B_COST_catalog * TRAIN_COST_SCALED
V2 = B_TIME * SM_TT_catalog + B_COST_catalog * SM_COST_SCALED
V3 = ASC_CAR_catalog + B_TIME * CAR_TT_catalog + B_COST_catalog * CAR_CO_SCALED

# %%
# Associate utility functions with the numbering of alternatives.
V = {1: V1, 2: V2, 3: V3}

# %%
# Associate the availability conditions with the alternatives.
av = {1: TRAIN_AV_SP, 2: SM_AV, 3: CAR_AV_SP}

# %%
# Definition of the model. This is the contribution of each
# observation to the log likelihood function.
logprob = models.loglogit(V, av, CHOICE)

# %%
print(
    f'Total number of possible specifications: '
    f'{logprob.number_of_multiple_expressions()}'
)

# %%
# Create the biogeme object.
the_biogeme = bio.BIOGEME(database, logprob)
the_biogeme.modelName = 'b21multiple_models'

# %%
# Name of the Pareto file.
PARETO_FILE_NAME = 'b21multiple_models.pareto'



"""

Mixtures of logit with Monte-Carlo 500 MLHS draws
=================================================

Estimation of a mixtures of logit models where the integral is
approximated using MonteCarlo integration with MLHS draws.

:author: Michel Bierlaire, EPFL
:date: Thu Apr 13 23:37:44 2023
"""

import biogeme.biogeme_logging as blog
from biogeme.expressions import bioDraws
from b07estimation_specification import get_biogeme

# %%
logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example b07estimation_monte_carlo_mlhs_500.py')

# %%
R = 500

the_draws = bioDraws('B_TIME_RND', 'NORMAL_MLHS')
the_biogeme = get_biogeme(the_draws=the_draws, number_of_draws=R)
the_biogeme.modelName = 'b07estimation_monte_carlo_mlhs_500'

# %%
results = the_biogeme.estimate()

# %%
print(results.short_summary())

# %%
pandas_results = results.getEstimatedParameters()
pandas_results



"""Arithmetic expressions accepted by Biogeme: binary max

Michel Bierlaire
Sat Jun 14 2025, 10:12:55
"""

from __future__ import annotations

import logging

import jax.numpy as jnp

from .base_expressions import ExpressionOrNumeric
from .binary_expressions import BinaryOperator
from .jax_utils import JaxFunctionType

logger = logging.getLogger(__name__)


class BinaryMax(BinaryOperator):
    """
    Maximum of two expressions
    """

    def __init__(self, left: ExpressionOrNumeric, right: ExpressionOrNumeric):
        """Constructor

        :param left: first arithmetic expression
        :type left: biogeme.expressions.Expression

        :param right: second arithmetic expression
        :type right: biogeme.expressions.Expression
        """
        super().__init__(left, right)

    def deep_flat_copy(self) -> BinaryMax:
        """Provides a copy of the expression. It is deep in the sense that it generates copies of the children.
        It is flat in the sense that any `MultipleExpression` is transformed into the currently selected expression.
        The flat part is irrelevant for this expression.
        """
        left_copy = self.left.deep_flat_copy()
        right_copy = self.right.deep_flat_copy()
        return type(self)(left=left_copy, right=right_copy)

    def __str__(self) -> str:
        return f'BinaryMax({self.left}, {self.right})'

    def __repr__(self) -> str:
        return f'BinaryMax({repr(self.left)}, {repr(self.right)})'

    def get_value(self) -> float:
        """Evaluates the value of the expression

        :return: value of the expression
        :rtype: float
        """
        if self.left.get_value() >= self.right.get_value():
            return self.left.get_value()

        return self.right.get_value()

    def recursive_construct_jax_function(
        self, numerically_safe: bool
    ) -> JaxFunctionType:
        """
        Generates a function to be used by biogeme_jax. Must be overloaded by each expression
        :return: the function takes two parameters: the parameters, and one row of the database.
        """
        left_jax: JaxFunctionType = self.left.recursive_construct_jax_function(
            numerically_safe=numerically_safe
        )
        right_jax: JaxFunctionType = self.right.recursive_construct_jax_function(
            numerically_safe=numerically_safe
        )

        def the_jax_function(
            parameters: jnp.ndarray,
            one_row: jnp.ndarray,
            the_draws: jnp.ndarray,
            the_random_variables: jnp.ndarray,
        ) -> jnp.ndarray:
            left_value = left_jax(parameters, one_row, the_draws, the_random_variables)
            right_value = right_jax(
                parameters, one_row, the_draws, the_random_variables
            )
            return jnp.maximum(left_value, right_value)

        return the_jax_function



"""
Test the distributions module

:author: Michel Bierlaire
:date: Sun Nov 19 19:52:22 2023

"""

import unittest
from biogeme.expressions import Numeric
from biogeme.distributions import (
    normalpdf,
    lognormalpdf,
    uniformpdf,
    triangularpdf,
    logisticcdf,
)


class TestNormalPdf(unittest.TestCase):
    def test_normal_values(self):
        # Test with normal values
        self.assertAlmostEqual(normalpdf(0).get_value(), 0.3989422804)
        self.assertAlmostEqual(normalpdf(1, 1, 1).get_value(), 0.3989422804)

    def test_expression_inputs(self):
        expression = normalpdf(1, 1, 1)
        value = expression.get_value()
        self.assertAlmostEqual(value, 0.3989422804)

    def test_negative_sigma(self):
        # Test with negative sigma (should handle or raise an error)
        with self.assertRaises(ValueError):
            normalpdf(0, 0, -1)

    def test_zero_sigma(self):
        # Test with sigma equal to zero (should handle or raise an error)
        with self.assertRaises(ValueError):
            normalpdf(0, 0, 0)

    def test_non_numeric_inputs(self):
        # Test with non-numeric inputs
        with self.assertRaises(TypeError):
            normalpdf("string", "string", "string")

    def test_mixed_inputs(self):
        # Test with mixed type inputs
        self.assertAlmostEqual(normalpdf(0.0, Numeric(0), 1).get_value(), 0.3989422804)
        # Add more mixed type tests here

    def test_large_values(self):
        # Test with very large values
        self.assertAlmostEqual(normalpdf(1000, 0, 1).get_value(), 0)

    def test_small_values(self):
        # Test with very small values
        self.assertAlmostEqual(normalpdf(0, 0, 0.0001).get_value(), 3989.4228034270454)


class TestLognormalPdf(unittest.TestCase):
    def test_normal_values(self):
        # Test with normal values
        self.assertAlmostEqual(lognormalpdf(1, 0, 1).get_value(), 0.3989422804)
        expected_value = 0.04222768901274423
        self.assertAlmostEqual(lognormalpdf(2, 3, 4).get_value(), expected_value)

    def test_expression_inputs(self):
        expression = lognormalpdf(1, 0, 1)
        value = expression.get_value()
        self.assertAlmostEqual(value, 0.3989422804)
        expression = lognormalpdf(2, 3, 4)
        value = expression.get_value()
        self.assertAlmostEqual(value, 0.04222768901274423)

    def test_invalid_x_values(self):
        # Test with invalid x values (x <= 0)
        with self.assertRaises(ValueError):
            lognormalpdf(0, 0, 1)
        with self.assertRaises(ValueError):
            lognormalpdf(-1, 0, 1)

    def test_negative_sigma(self):
        # Test with negative sigma (should handle or raise an error)
        with self.assertRaises(ValueError):
            lognormalpdf(1, 0, -1)

    def test_zero_sigma(self):
        # Test with sigma equal to zero (should handle or raise an error)
        with self.assertRaises(ValueError):
            lognormalpdf(1, 0, 0)

    def test_non_numeric_inputs(self):
        # Test with non-numeric inputs
        with self.assertRaises(TypeError):
            lognormalpdf("string", "string", "string")

    def test_mixed_inputs(self):
        # Test with mixed type inputs
        self.assertAlmostEqual(
            lognormalpdf(1.0, Numeric(0), 1).get_value(), 0.3989422804
        )
        # Add more mixed type tests here

    def test_large_values(self):
        # Test with very large values
        self.assertAlmostEqual(
            lognormalpdf(1000, 0, 1).get_value(), 0
        )  # Replace with actual expected value

    def test_small_values(self):
        # Test with very small but positive values
        expected_value = 0
        self.assertAlmostEqual(
            lognormalpdf(0.0001, 0, 1).get_value(), expected_value
        )  # Replace with actual expected value


class TestUniformPdf(unittest.TestCase):
    def test_within_bounds(self):
        # Test with x within the bounds [a, b]
        self.assertAlmostEqual(uniformpdf(0, -1, 1).get_value(), 0.5)
        self.assertAlmostEqual(uniformpdf(0.5, -1, 1).get_value(), 0.5)

    def test_outside_bounds(self):
        # Test with x outside the bounds [a, b]
        self.assertAlmostEqual(uniformpdf(-2, -1, 1).get_value(), 0.0)
        self.assertAlmostEqual(uniformpdf(2, -1, 1).get_value(), 0.0)

    def test_expression_inputs(self):
        self.assertAlmostEqual(
            uniformpdf(Numeric(0), Numeric(-1), Numeric(1)).get_value(), 0.5
        )
        self.assertAlmostEqual(
            uniformpdf(Numeric(0.5), Numeric(-1), Numeric(1)).get_value(), 0.5
        )

    def test_invalid_bounds(self):
        # Test with invalid bounds (a >= b)
        with self.assertRaises(ValueError):
            uniformpdf(0, 1, -1)

    def test_non_numeric_inputs(self):
        # Test with non-numeric inputs
        with self.assertRaises(TypeError):
            uniformpdf("string", "string", "string")

    def test_mixed_inputs(self):
        # Test with mixed type inputs
        self.assertAlmostEqual(uniformpdf(0.5, Numeric(-1), 1).get_value(), 0.5)

    def test_edge_cases(self):
        # Test edge cases (x at the boundaries a and b)
        self.assertAlmostEqual(uniformpdf(-1, -1, 1).get_value(), 0.5)
        self.assertAlmostEqual(uniformpdf(1, -1, 1).get_value(), 0.5)


class TestTriangularPdf(unittest.TestCase):
    def test_valid_values(self):
        # Test with x within the bounds [a, b] and c within (a, b)
        self.assertAlmostEqual(triangularpdf(0, -1, 1, 0).get_value(), 1.0)
        self.assertAlmostEqual(triangularpdf(-0.5, -1, 1, 0).get_value(), 0.5)
        self.assertAlmostEqual(triangularpdf(0.5, -1, 1, 0).get_value(), 0.5)

    def test_outside_bounds(self):
        # Test with x outside the bounds [a, b]
        self.assertAlmostEqual(triangularpdf(-2, -1, 1, 0).get_value(), 0.0)
        self.assertAlmostEqual(triangularpdf(2, -1, 1, 0).get_value(), 0.0)

    def test_invalid_c_values(self):
        # Test with invalid c values (c <= a or c >= b)
        with self.assertRaises(ValueError):
            triangularpdf(0, -1, 1, -1)
        with self.assertRaises(ValueError):
            triangularpdf(0, -1, 1, 2)

    def test_expression_inputs(self):
        # Test with Expression type inputs
        self.assertAlmostEqual(
            triangularpdf(Numeric(0), Numeric(-1), Numeric(1), Numeric(0)).get_value(),
            1.0,
        )
        self.assertAlmostEqual(
            triangularpdf(
                Numeric(-0.5), Numeric(-1), Numeric(1), Numeric(0)
            ).get_value(),
            0.5,
        )
        self.assertAlmostEqual(
            triangularpdf(
                Numeric(0.5), Numeric(-1), Numeric(1), Numeric(0)
            ).get_value(),
            0.5,
        )

    def test_non_numeric_inputs(self):
        # Test with non-numeric inputs
        with self.assertRaises(TypeError):
            triangularpdf("string", "string", "string", "string")

    def test_mixed_inputs(self):
        # Test with mixed type inputs
        self.assertAlmostEqual(triangularpdf(0.5, Numeric(-1), 1, 0).get_value(), 0.5)
        # Add more mixed type tests here

    def test_edge_cases(self):
        # Test edge cases (x at the boundaries a, b, and c)
        self.assertAlmostEqual(triangularpdf(-1, -1, 1, 0).get_value(), 0.0)
        self.assertAlmostEqual(triangularpdf(1, -1, 1, 0).get_value(), 0.0)
        self.assertAlmostEqual(triangularpdf(0, -1, 1, 0).get_value(), 1.0)


class TestLogisticCdf(unittest.TestCase):
    def test_normal_values(self):
        # Test with normal values
        self.assertAlmostEqual(logisticcdf(0).get_value(), 0.5)
        self.assertAlmostEqual(logisticcdf(1, 0, 1).get_value(), 0.7310585786300049)

    def test_expression_inputs(self):
        # Test with Expression type inputs if applicable
        self.assertAlmostEqual(logisticcdf(Numeric(0)).get_value(), 0.5)
        self.assertAlmostEqual(
            logisticcdf(Numeric(1), Numeric(0), Numeric(1)).get_value(),
            0.7310585786300049,
        )

    def test_negative_scale(self):
        # Test with negative scale (should handle or raise an error)
        with self.assertRaises(ValueError):
            logisticcdf(0, 0, -1)

    def test_zero_scale(self):
        # Test with scale equal to zero (should handle or raise an error)
        with self.assertRaises(ValueError):
            logisticcdf(0, 0, 0)

    def test_non_numeric_inputs(self):
        # Test with non-numeric inputs
        with self.assertRaises(TypeError):
            logisticcdf("string", "string", "string")

    def test_mixed_inputs(self):
        # Test with mixed type inputs
        self.assertAlmostEqual(logisticcdf(0.0, Numeric(0), 1).get_value(), 0.5)
        # Add more mixed type tests here

    def test_large_values(self):
        # Test with very large values
        self.assertAlmostEqual(logisticcdf(1000, 0, 1).get_value(), 1.0)

    def test_small_values(self):
        # Test with very small values
        self.assertAlmostEqual(logisticcdf(-1000, 0, 1).get_value(), 0.0)


if __name__ == '__main__':
    unittest.main()



import warnings
from .biogeme_logging import *  # Import everything from the new module

# Issue the deprecation warning
warnings.warn(
    "The module 'biogeme.logging' is deprecated and will be removed in a future version. "
    "Please use 'biogeme.biogeme_logging' instead. This is due to a name conflict "
    "with the main logging module.",
    DeprecationWarning,
    stacklevel=2,
)



"""

Measurement equations: discrete indicators
==========================================

Ordered probit.

:author: Michel Bierlaire, EPFL
:date: Fri Apr 14 09:39:10 2023

"""

import biogeme.biogeme_logging as blog
import biogeme.biogeme as bio
from biogeme.expressions import (
    Beta,
    log,
    Elem,
    bioNormalCdf,
    Variable,
    bioMultSum,
)

from optima import (
    database,
    male,
    age,
    haveChildren,
    highEducation,
    childCenter,
    childSuburb,
    SocioProfCat,
)

logger = blog.get_screen_logger(level=blog.INFO)
logger.info('Example m01_latent_variable.py')

# %%
# Parameters for the structural equation
coef_intercept = Beta('coef_intercept', 0.0, None, None, 0)
coef_age_30_less = Beta('coef_age_30_less', 0.0, None, None, 0)
coef_male = Beta('coef_male', 0.0, None, None, 0)
coef_haveChildren = Beta('coef_haveChildren', 0.0, None, None, 0)
coef_highEducation = Beta('coef_highEducation', 0.0, None, None, 0)
coef_artisans = Beta('coef_artisans', 0.0, None, None, 0)
coef_employees = Beta('coef_employees', 0.0, None, None, 0)
coef_child_center = Beta('coef_child_center', 0.0, None, None, 0)
coef_child_suburb = Beta('coef_child_suburb', 0.0, None, None, 0)

# %%
# Latent variable: structural equation
ACTIVELIFE = (
    coef_intercept
    + coef_child_center * childCenter
    + coef_child_suburb * childSuburb
    + coef_highEducation * highEducation
    + coef_artisans * (SocioProfCat == 5)
    + coef_employees * (SocioProfCat == 6)
    + coef_age_30_less * (age <= 30)
    + coef_male * male
    + coef_haveChildren * haveChildren
)

# %%
# Measurement equations
indicators = [
    'ResidCh01',
    'ResidCh04',
    'ResidCh05',
    'ResidCh06',
    'LifSty07',
    'LifSty10',
]

# %%
# We define the intercept parameters. The first one is normalized to 0.
inter = {k: Beta(f'inter_{k}', 0, None, None, 0) for k in indicators[1:]}
inter[indicators[0]] = Beta(f'INTER_{indicators[0]}', 0, None, None, 1)

# %%
# We define the coefficients. The first one is normalized to 1.
coefficients = {k: Beta(f'coeff_{k}', 0, None, None, 0) for k in indicators[1:]}
coefficients[indicators[0]] = Beta(f'B_{indicators[0]}', 1, None, None, 1)

# %%
# We define the measurement equations for each indicator
models = {k: inter[k] + coefficients[k] * ACTIVELIFE for k in indicators}

# %%
# We define the scale parameters of the error terms.
sigma_star = {k: Beta(f'sigma_star_{k}', 1, 1.0e-5, None, 0) for k in indicators[1:]}
sigma_star[indicators[0]] = Beta(f'sigma_star_{indicators[0]}', 1, None, None, 1)

# %%
# Symmetric threshold.
delta_1 = Beta('delta_1', 0.1, 1.0e-5, None, 0)
delta_2 = Beta('delta_2', 0.2, 1.0e-5, None, 0)
tau_1 = -delta_1 - delta_2
tau_2 = -delta_1
tau_3 = delta_1
tau_4 = delta_1 + delta_2

# %%
# Ordered probit models.
tau_1_residual = {k: (tau_1 - models[k]) / sigma_star[k] for k in indicators}
tau_2_residual = {k: (tau_2 - models[k]) / sigma_star[k] for k in indicators}
tau_3_residual = {k: (tau_3 - models[k]) / sigma_star[k] for k in indicators}
tau_4_residual = {k: (tau_4 - models[k]) / sigma_star[k] for k in indicators}
dict_prob_indicators = {
    k: {
        1: bioNormalCdf(tau_1_residual[k]),
        2: bioNormalCdf(tau_2_residual[k]) - bioNormalCdf(tau_1_residual[k]),
        3: bioNormalCdf(tau_3_residual[k]) - bioNormalCdf(tau_2_residual[k]),
        4: bioNormalCdf(tau_4_residual[k]) - bioNormalCdf(tau_3_residual[k]),
        5: 1 - bioNormalCdf(tau_4_residual[k]),
        6: 1.0,
        -1: 1.0,
        -2: 1.0,
    }
    for k in indicators
}

# %%
log_proba = {k: log(Elem(dict_prob_indicators[k], Variable(k))) for k in indicators}
loglike = bioMultSum(log_proba)

# %%
# Create the Biogeme object
biogeme = bio.BIOGEME(database, loglike)
biogeme.modelName = 'm01_latent_variable'

# %%
# Estimate the parameters
results = biogeme.estimate()

# %%
results.getEstimatedParameters()



"""Defines  a catalog containing generic and alternative specific specifications

Michel Bierlaire
Sun Apr 27 2025, 15:49:25
"""

from biogeme.exceptions import BiogemeError
from biogeme.expressions import Beta
from biogeme.segmentation import DiscreteSegmentationTuple
from .catalog import Catalog
from .controller import Controller
from .segmentation_catalog import SegmentedParameters, segmentation_catalogs


def generic_alt_specific_catalogs(
    generic_name: str,
    beta_parameters: list[Beta],
    alternatives: tuple[str, ...],
    potential_segmentations: tuple[DiscreteSegmentationTuple, ...] | None = None,
    maximum_number=5,
):
    """Generate catalogs selecting generic or alternative specific coefficients

    :param generic_name: name associated with all the parameters in the catalog
    :type generic_name: str

    :param beta_parameters: coefficients of interest
    :type beta_parameters: list(biogeme.expressions.Beta)

    :param alternatives: names of the alternatives
    :type alternatives: tuple(str)

    :param potential_segmentations: tuple of potential segmentations, or None
    :type potential_segmentations: tuple(biogeme.segmentation.DiscreteSegmentationTuple)

    :param maximum_number: maximum number of segmentations to consider
    :type maximum_number: int

    :return: a list of catalogs for each alternative
    :rtype: list(dict(str: biogeme.catalog.Catalog))
    """
    if len(alternatives) < 2:
        error_msg = (
            f'An alternative specific specification requires at least 2 '
            f'alternatives, and not {len(alternatives)}'
        )
        raise BiogemeError(error_msg)

    if not isinstance(beta_parameters, list):
        error_msg = (
            f'Argument "beta_parameters" of function '
            f'"{generic_alt_specific_catalogs.__name__}" must be a list.'
        )
        raise BiogemeError(error_msg)

    wrong_indices = []
    for index, beta in enumerate(beta_parameters):
        if not isinstance(beta, Beta):
            wrong_indices.append(index)

    if wrong_indices:
        error_msg = (
            f'The entries at the following indices are not Beta expressions: '
            f'{wrong_indices}'
        )
        raise BiogemeError(error_msg)

    # We first generate the alternative specific versions of the parameters
    generic_parameters = beta_parameters
    the_segmented_parameters = SegmentedParameters(
        beta_parameters=generic_parameters,
        alternatives=alternatives,
    )

    # If applicable, we apply the potential segmentations
    if potential_segmentations:
        segmented_catalogs = segmentation_catalogs(
            generic_name=generic_name,
            beta_parameters=the_segmented_parameters.all_parameters,
            potential_segmentations=potential_segmentations,
            maximum_number=maximum_number,
        )

    def get_expression(param_index: int, alternative: str | None):
        """Returns either the parameter, or the segmented version if applicable"""

        if potential_segmentations:
            the_index = the_segmented_parameters.get_index(param_index, alternative)
            return segmented_catalogs[the_index]
        return the_segmented_parameters.get_beta(param_index, alternative)

    # We now control for generic or alternative specific with a single
    # controller for all catalogs
    the_controller = Controller(
        controller_name=f'{generic_name}_gen_altspec',
        specification_names=('generic', 'altspec'),
    )

    # We organize the catalogs as a list of dict
    results = []
    for index, beta in enumerate(beta_parameters):
        the_dict = {
            alternative: Catalog.from_dict(
                catalog_name=f'{beta.name}_{alternative}_gen_altspec',
                dict_of_expressions={
                    'generic': get_expression(index, None),
                    'altspec': get_expression(index, alternative),
                },
                controlled_by=the_controller,
            )
            for alternative in alternatives
        }
        results.append(the_dict)
    return results



""" Module in charge of functionalities related to the sampling of alternatives

:author: Michel Bierlaire
:date: Thu Sep  7 10:14:54 2023
"""

import copy
import logging

import numpy as np
import pandas as pd

from biogeme.exceptions import BiogemeError
from .sampling_context import SamplingContext, LOG_PROBA_COL, MEV_WEIGHT, CNL_PREFIX

logger = logging.getLogger(__name__)


def generate_segment_size(sample_size: int, number_of_segments: int) -> list[int]:
    """This function calculates the size of each segment, so that
    they are as close to each other as possible, and cover the full sample

    :param sample_size: total size of the sample
    :type sample_size: int

    :param number_of_segments: number of segments
    :type number_of_segments: int

    :return: list of length number_of_segments, containing the segment sizes
    :rtype: list[int]

    """
    if sample_size < 0:
        raise ValueError("Sample size cannot be negative.")
    if number_of_segments <= 0:
        raise ValueError("Number of segments must be positive.")

    # Calculate the base value and the remainder
    base_value = sample_size // number_of_segments
    remainder = sample_size % number_of_segments

    # Distribute the base value across the list
    segment_sizes = [base_value] * number_of_segments

    # Distribute the remainder across the first few elements
    for i in range(remainder):
        segment_sizes[i] += 1

    return segment_sizes


class SamplingOfAlternatives:
    """Class dealing with the various methods needed to estimate
    models with samples of alternatives

    """

    def __init__(self, context: SamplingContext):
        """Constructor

        :param context: contains all the information that is needed to
            perform the sampling of alternatives.

        """

        self.alternatives = context.alternatives
        self.id_column = context.id_column
        self.partition = context.sampling_protocol
        self.second_partition = context.mev_sampling_protocol
        self.cnl_nests = context.cnl_nests

    def add_alphas(self, the_sample: pd.DataFrame) -> pd.DataFrame:
        """Add the alpha parameters in the sampled database"""

        if self.cnl_nests is None:
            raise BiogemeError(
                f'No nests have been defined for the cross-nested logit model'
            )

        # We add the alpha parameters in the sample
        def get_alphas(alternative_id: int) -> pd.Series:
            """Prepare the alphas for insertion in the data frame"""
            assert self.cnl_nests is not None
            the_dict = self.cnl_nests.get_alpha_values(alternative_id)
            return pd.Series(the_dict)

        new_columns = the_sample[self.id_column].apply(get_alphas)
        new_columns = new_columns.rename(columns=lambda x: CNL_PREFIX + x)

        the_sample = pd.concat([the_sample, new_columns], axis="columns")
        return the_sample

    def sample_alternatives(self, chosen: int) -> pd.DataFrame:
        """Performing the sampling of alternatives

        :param chosen: ID of the chosen alternative, that must be included
            in the choice set.

        :return: data frame containing a sample of
            alternatives. The first one is the chosen alternative
        :raise BiogemeError: if the chosen alternative is unknown.

        """
        chosen_alternative = self.alternatives[
            self.alternatives[self.id_column] == chosen
        ].copy()
        if len(chosen_alternative) < 1:
            error_msg = f"Unknown alternative: {chosen}"
            raise BiogemeError(error_msg)
        if len(chosen_alternative) > 1:
            error_msg = f"Duplicate alternative: {chosen}"
            raise BiogemeError(error_msg)

        results = []

        for stratum in self.partition:
            # stratum.subset is a set of int
            # We create a copy because we'll have to drop the chosen alternative
            the_subset_of_alternatives = copy.deepcopy(stratum.subset)
            stratum_size = len(stratum.subset)
            sample_size = stratum.sample_size
            logproba = np.log(sample_size) - np.log(stratum_size)
            if chosen in stratum.subset:
                # Discard the chosen alternative
                the_subset_of_alternatives.discard(chosen)
                # And we sample one alternative less
                sample_size -= 1
                # Include the correction terms
                chosen_alternative[LOG_PROBA_COL] = logproba

            # subset is a pandas data frame containing the description
            # of all alternatives in the subset
            subset = self.alternatives[
                self.alternatives[self.id_column].isin(the_subset_of_alternatives)
            ]
            # Perform the sampling
            sample = subset.sample(
                n=sample_size, replace=False, axis="index", ignore_index=True
            )

            sample[LOG_PROBA_COL] = logproba

            results.append(sample)

        the_sample = pd.concat(results, ignore_index=True)
        # Add the chosen alternative. By construction, it is not in the sample.
        the_sample = pd.concat([chosen_alternative, the_sample], ignore_index=True)
        if self.cnl_nests:
            self.add_alphas(the_sample)
        return the_sample

    def sample_mev_alternatives(self) -> pd.DataFrame:
        """Performing the sampling of alternatives for the MEV
        terms. Here, the chosen alternative is ignored.

        :return: data frame containing a sample of alternatives

        """
        results = []

        for stratum in self.second_partition:
            stratum_size = len(stratum.subset)
            sample_size = stratum.sample_size
            mev_weight = stratum_size / sample_size
            subset = self.alternatives[
                self.alternatives[self.id_column].isin(stratum.subset)
            ]
            sample = subset.sample(
                n=sample_size, replace=False, axis="index", ignore_index=True
            )
            sample[MEV_WEIGHT] = mev_weight

            results.append(sample)

        the_sample = pd.concat(results, ignore_index=True)

        if self.cnl_nests:
            self.add_alphas(the_sample)

        return the_sample



"""Arithmetic expressions accepted by Biogeme: ConditionalSum

Michel Bierlaire
Sat Sep  9 15:29:36 2023
"""

from __future__ import annotations

import logging
from typing import Iterable, NamedTuple

import jax
from biogeme.exceptions import BiogemeError

from .base_expressions import Expression, ExpressionOrNumeric
from .convert import validate_and_convert
from .jax_utils import JaxFunctionType

logger = logging.getLogger(__name__)


class ConditionalTermTuple(NamedTuple):
    condition: ExpressionOrNumeric
    term: ExpressionOrNumeric


class ConditionalSum(Expression):
    """This expression returns the sum of a selected list of
    expressions. An expression is considered in the sum only if the
    corresponding key is True (that is, return a non-zero value).


    """

    def __init__(self, list_of_terms: Iterable[ConditionalTermTuple]):
        """Constructor

        :param list_of_terms: list containing the terms and the associated conditions

        :raise BiogemeError: if one of the expressions is invalid, that is
            neither a numeric value nor a
            biogeme.expressions.Expression object.
        :raise BiogemeError: if the dict of expressions is empty
        :raise BiogemeError: if the dict of expressions is not a dict

        """
        if not list_of_terms:
            raise BiogemeError('The argument of ConditionalSum cannot be empty')

        super().__init__()

        self.list_of_terms = [
            the_term._replace(
                condition=validate_and_convert(the_term.condition),
                term=validate_and_convert(the_term.term),
            )
            for the_term in list_of_terms
        ]
        for the_term in self.list_of_terms:
            self.children.append(the_term.condition)
            self.children.append(the_term.term)

    def deep_flat_copy(self) -> ConditionalSum:
        """Provides a copy of the expression. It is deep in the sense that it generates copies of the children.
        It is flat in the sense that any `MultipleExpression` is transformed into the currently selected expression.
        The flat part is irrelevant for this expression.
        """
        copy_list_of_terms: list[ConditionalTermTuple] = [
            ConditionalTermTuple(
                condition=the_term.condition.deep_flat_copy(),
                term=the_term.term.deep_flat_copy(),
            )
            for the_term in self.list_of_terms
        ]
        return type(self)(list_of_terms=copy_list_of_terms)

    def get_value(self) -> float:
        """Evaluates the value of the expression

        :return: value of the expression
        :rtype: float
        """
        result = 0.0
        for the_term in self.list_of_terms:
            condition = the_term.condition.get_value()
            if condition != 0:
                result += the_term.term.get_value()
        return result

    def recursive_construct_jax_function(
        self, numerically_safe: bool
    ) -> JaxFunctionType:
        compiled_terms = [
            (
                cond.recursive_construct_jax_function(
                    numerically_safe=numerically_safe
                ),
                term.recursive_construct_jax_function(
                    numerically_safe=numerically_safe
                ),
            )
            for cond, term in self.list_of_terms
        ]

        def the_jax_function(parameters, one_row, the_draws, the_random_variables):
            result = 0.0
            for cond_fn, term_fn in compiled_terms:
                cond_val = cond_fn(parameters, one_row, the_draws, the_random_variables)

                def include_branch(_):
                    return term_fn(parameters, one_row, the_draws, the_random_variables)

                def skip_branch(_):
                    return 0.0

                result += jax.lax.cond(
                    cond_val != 0.0, include_branch, skip_branch, operand=None
                )

            return result

        return the_jax_function

    def __str__(self) -> str:
        s = (
            'ConditionalSum('
            + ', '.join([f'{k}: {v}' for k, v in self.list_of_terms])
            + ')'
        )
        return s

    def __repr__(self) -> str:
        return f"ConditionalSum({repr(self.list_of_terms)})"



"""

True parameters
===============

Value of the parameters used to generate synthetic choices.
"""

# %%
SCALE = 0.5
true_parameters = {
    'beta_rating': 1.5 * SCALE,
    'beta_price': -0.8 * SCALE,
    'beta_chinese': 1.5 * SCALE,
    'beta_japanese': 2.5 * SCALE,
    'beta_korean': 1.5 * SCALE,
    'beta_indian': 2 * SCALE,
    'beta_french': 1.5 * SCALE,
    'beta_mexican': 2.5 * SCALE,
    'beta_lebanese': 1.5 * SCALE,
    'beta_ethiopian': 1 * SCALE,
    'beta_log_dist': -1.2 * SCALE,
    'mu_asian': 2.0,
    'mu_downtown': 2.0,
}



"""
Test the parameters module

:author: Michel Bierlaire
:date: Tue Nov 29 10:26:32 2022

"""

import shutil
from os import path
import tempfile
import unittest
import numpy as np
from biogeme.parameters import Parameters
import biogeme.exceptions as excep
from biogeme.default_parameters import ParameterTuple
import biogeme.check_parameters as cp
import biogeme.optimization as opt

BLANKS = 25 * ' '

FILE_CONTENT = """
[Specification]
suggest_scales = "True" # bool: If True, Biogeme suggests the scaling of
                         # the variables in the database.
missing_data = 99999 # number: If one variable has this value, it is assumed
                         # that a data is missing and an exception will
                         # be triggered.

[Output]
generate_html = "True" # bool: "True" if the HTML file
                         # with the results must be generated.
generate_pickle = "True" # bool: "True" if the pickle file
                         # with the results must be generated.

[Estimation]
second_derivatives = 1.0 # float: proportion (between 0 and 1) of iterations when the analytical Hessian is calculated
tolerance = 6.06273418136464e-06 # float: the algorithm stops when this precision is reached
max_iterations = 100 # int: maximum number of iterations
optimization_algorithm = "simple_bounds" # str: optimization algorithm to be used for estimation.
                         # Valid values: ['scipy', 'LS-newton', 'TR-newton', 'LS-BFGS', 'TR-BFGS', 'simple_bounds']

save_iterations = "True" # bool: If True, the current iterate is saved after
                         # each iteration, in a file named
                         # ``__[modelName].iter``, where
                         # ``[modelName]`` is the name given to the
                         # model. If such a file exists, the starting
                         # values for the estimation are replaced by
                         # the values saved in the file.

a_param = "lopsum"

[MonteCarlo]
number_of_draws = 1000 # int: Number of draws for Monte-Carlo integration.
seed = 0 # int: Seed used for the pseudo-random number generation. It is
                         # useful only when each run should generate the exact same
                         # result. If 0, a new seed is used at each run.

[MultiThreading]
number_of_threads = 0 # int: Number of threads/processors to be used. If
                         # the parameter is 0, the number of
                         # available threads is calculated using
                         # cpu_count().

"""


default_parameters = (
    ParameterTuple(
        name='second_derivatives',
        value=1.0,
        type=float,
        section='Estimation',
        description='# float: proportion (between 0 and 1) of iterations when the analytical Hessian is calculated',
        check=(cp.zero_one, cp.is_number),
    ),
    ParameterTuple(
        name='tolerance',
        value=np.finfo(np.float64).eps ** 0.3333,
        type=float,
        section='Estimation',
        description='# float: the algorithm stops when this precision is reached',
        check=(cp.zero_one, cp.is_number),
    ),
    ParameterTuple(
        name='max_iterations',
        value=100,
        type=int,
        section='Estimation',
        description='# int: maximum number of iterations',
        check=(cp.is_integer, cp.is_positive),
    ),
    ParameterTuple(
        name='optimization_algorithm',
        value='simple_bounds',
        type=str,
        section='Estimation',
        description=(
            f'# str: optimization algorithm to be used for estimation.\n'
            f'{BLANKS}# Valid values: {list(opt.algorithms.keys())}'
        ),
        check=(cp.check_algo_name,),
    ),
    ParameterTuple(
        name='generate_html',
        value=True,
        type=bool,
        section='Output',
        description=(
            '# bool: "True" if the HTML file\n'
            '{BLANKS}# with the results must be generated.'
        ),
        check=(cp.is_boolean,),
    ),
    ParameterTuple(
        name='generate_pickle',
        value=True,
        type=bool,
        section='Output',
        description=(
            f'# bool: "True" if the pickle file\n'
            f'{BLANKS}# with the results must be generated.'
        ),
        check=(cp.is_boolean,),
    ),
    ParameterTuple(
        name='number_of_threads',
        value=0,
        type=int,
        section='MultiThreading',
        description=(
            f'# int: Number of threads/processors to be used. If\n'
            f'{BLANKS}# the parameter is 0, the number of\n'
            f'{BLANKS}# available threads is calculated using\n'
            f'{BLANKS}# cpu_count().'
        ),
        check=(cp.is_integer, cp.is_non_negative),
    ),
    ParameterTuple(
        name='number_of_draws',
        value=1000,
        type=int,
        section='MonteCarlo',
        description=('int: Number of draws for Monte-Carlo integration.'),
        check=(cp.is_integer, cp.is_positive),
    ),
    ParameterTuple(
        name='skip_audit',
        value=False,
        type=bool,
        section='Specification',
        description=(
            f'# bool: If True, does not check the validity of the\n'
            f'{BLANKS}# formulas. It may save significant amount of\n'
            f'{BLANKS}# time for large models and large data sets.\n'
        ),
        check=(cp.is_boolean,),
    ),
    ParameterTuple(
        name='suggest_scales',
        value=True,
        type=bool,
        section='Specification',
        description=(
            f'# bool: If True, Biogeme suggests the scaling of\n'
            f'{BLANKS}# the variables in the database.'
        ),
        check=(cp.is_boolean,),
    ),
    ParameterTuple(
        name='missing_data',
        value=99999,
        type=int,
        section='Specification',
        description=(
            f'# number: If one variable has this value, it is assumed\n'
            f'{BLANKS}# that a data is missing and an exception will\n'
            f'{BLANKS}# be triggered.'
        ),
        check=(cp.is_number,),
    ),
    ParameterTuple(
        name='seed',
        value=0,
        type=int,
        section='MonteCarlo',
        description=(
            f'# int: Seed used for the pseudo-random number generation. It is\n'
            f'{BLANKS}# useful only when each run should generate the exact same\n'
            f'{BLANKS}# result. If 0, a new seed is used at each run.'
        ),
        check=(cp.is_integer, cp.is_non_negative),
    ),
    ParameterTuple(
        name='save_iterations',
        value=True,
        type=bool,
        section='Estimation',
        description=(
            f'# bool: If True, the current iterate is saved after\n'
            f'{BLANKS}# each iteration, in a file named\n'
            f'{BLANKS}# ``__[modelName].iter``, where\n'
            f'{BLANKS}# ``[modelName]`` is the name given to the\n'
            f'{BLANKS}# model. If such a file exists, the starting\n'
            f'{BLANKS}# values for the estimation are replaced by\n'
            f'{BLANKS}# the values saved in the file.'
        ),
        check=(cp.is_boolean,),
    ),
    ParameterTuple(
        name='a_param',
        value='lopsum',
        type=str,
        section='Estimation',
        description='',
        check=None,
    ),
    ParameterTuple(
        name='a_param',
        value='anything',
        type=str,
        section='Specification',
        description='',
        check=None,
    ),
)


class TestParameters(unittest.TestCase):
    def setUp(self):
        self.biogeme_parameters = Parameters()
        for p in default_parameters:
            self.biogeme_parameters.add_parameter(p)

    def test_get_param_tuple_1(self):
        """Test the retrieval of parameter without the section name"""
        the_tuple = self.biogeme_parameters.get_param_tuple(name='missing_data')
        self.assertEqual(the_tuple.name, 'missing_data')
        self.assertEqual(the_tuple.value, 99999)
        self.assertEqual(the_tuple.section, 'Specification')

    def test_get_param_tuple_2(self):
        """Test the retrieval of parameter with the section name"""
        the_tuple = self.biogeme_parameters.get_param_tuple(
            name='missing_data', section='Specification'
        )
        self.assertEqual(the_tuple.name, 'missing_data')
        self.assertEqual(the_tuple.value, 99999)
        self.assertEqual(the_tuple.section, 'Specification')

    def test_get_param_tuple_3(self):
        """Test the error if the parameter does not appear in the
        mentioned section
        """
        with self.assertRaises(excep.BiogemeError):
            _ = self.biogeme_parameters.get_param_tuple(
                name='missing_data', section='Estimation'
            )

        with self.assertRaises(excep.BiogemeError):
            _ = self.biogeme_parameters.get_param_tuple(name='unknown_parameter')

        with self.assertRaises(excep.BiogemeError):
            _ = self.biogeme_parameters.get_param_tuple(
                name='unknown_parameter', section='Estimation'
            )

    def test_get_param_tuple_4(self):
        """Test the error if the parameter does appears in two
        sections, and the section is not specified.
        """
        with self.assertRaises(excep.BiogemeError):
            _ = self.biogeme_parameters.get_param_tuple(
                name='a_param',
            )

    def test_get_param_tuple_5(self):
        """Test the retrieval of a parameter that appears in two
        sections, with the section name

        """
        the_tuple = self.biogeme_parameters.get_param_tuple(
            name='a_param', section='Estimation'
        )
        self.assertEqual(the_tuple.name, 'a_param')
        self.assertEqual(the_tuple.value, 'lopsum')
        self.assertEqual(the_tuple.section, 'Estimation')

    def test_get_param_tuple_6(self):
        """Test the retrieval of a parameter that appears in two
        section, with the section name

        """
        the_tuple = self.biogeme_parameters.get_param_tuple(
            name='a_param', section='Specification'
        )
        self.assertEqual(the_tuple.name, 'a_param')
        self.assertEqual(the_tuple.value, 'anything')
        self.assertEqual(the_tuple.section, 'Specification')

    def test_check_parameter_value(self):
        a_tuple = ParameterTuple(
            name='param_name',
            value=2.0,
            type=float,
            section='Estimation',
            description='# float: proportion (between 0 and 1) of iterations when the analytical Hessian is calculated',
            check=(cp.zero_one, cp.is_number),
        )
        ok, _ = self.biogeme_parameters.check_parameter_value(a_tuple)
        self.assertFalse(ok)

    def test_add_parameter(self):
        a_tuple = ParameterTuple(
            name='param_name',
            value=0.5,
            type=float,
            section='Estimation',
            description='# float: proportion (between 0 and 1) of iterations when the analytical Hessian is calculated',
            check=(cp.zero_one, cp.is_number),
        )
        self.biogeme_parameters.add_parameter(a_tuple)
        self.assertEqual(
            self.biogeme_parameters.get_value(
                name=a_tuple.name, section=a_tuple.section
            ),
            0.5,
        )
        another_tuple = ParameterTuple(
            name='param_name',
            value=1.5,
            type=float,
            section='Estimation',
            description='# float: proportion (between 0 and 1) of iterations when the analytical Hessian is calculated',
            check=(cp.zero_one, cp.is_number),
        )
        with self.assertRaises(excep.BiogemeError):
            self.biogeme_parameters.add_parameter(another_tuple)

    def test_set_get_value(self):
        value = 1.0e-3
        self.biogeme_parameters.set_value('tolerance', value, section='Estimation')
        check = self.biogeme_parameters.get_value('tolerance', section='Estimation')
        self.assertEqual(value, check)


class TestToml(unittest.TestCase):
    """Tests for the biogeme.toml module"""

    def setUp(self):
        self.biogeme_parameters = Parameters()
        for p in default_parameters:
            self.biogeme_parameters.add_parameter(p)

    def test_from_file(self):
        """Read the parameters from file"""
        # Create a temporary directory
        test_dir = tempfile.mkdtemp()
        test_file = path.join(test_dir, 'biogeme.toml')
        with open(test_file, 'w', encoding='utf-8') as f:
            print(FILE_CONTENT, file=f)
        self.biogeme_parameters.read_file(file_name=test_file)

        with self.assertRaises(excep.BiogemeError):
            _ = self.biogeme_parameters.get_value('a_wrong_param', section='Estimation')

        check_int = self.biogeme_parameters.get_value('missing_data')
        self.assertEqual(check_int, 99999)

        check_float = self.biogeme_parameters.get_value(
            'second_derivatives', section='Estimation'
        )
        self.assertEqual(check_float, 1.0)

        check_str = self.biogeme_parameters.get_value('optimization_algorithm')
        self.assertEqual(check_str, 'simple_bounds')

        # Remove the directory after the test
        shutil.rmtree(test_dir)

    def test_from_default(self):
        """Test with default parameters"""
        test_dir = tempfile.mkdtemp()
        test_file = path.join(test_dir, 'any_file_name.toml')
        self.biogeme_parameters.read_file(file_name=test_file)

        with self.assertRaises(excep.BiogemeError):
            _ = self.biogeme_parameters.get_value('a_wrong_param', section='Estimation')

        check_int = self.biogeme_parameters.get_value('missing_data')
        self.assertEqual(check_int, 99999)

        check_float = self.biogeme_parameters.get_value(
            'second_derivatives', section='Estimation'
        )
        self.assertEqual(check_float, 1.0)

        check_str = self.biogeme_parameters.get_value('optimization_algorithm')
        self.assertEqual(check_str, 'simple_bounds')
        # Remove the directory after the test
        shutil.rmtree(test_dir)


if __name__ == '__main__':
    unittest.main()
